#include "common.h"





// ########################################################################
//
// Constants
// =========
//
// ########################################################################





// ########################################################################
//
// Variables
// =========
//
// ########################################################################





bool NoInfoMessages = false;
bool NoErrorMessages = false;
bool ProcessingStarted = false;

wchar_t InstancePath [MAX_PATH];





// ########################################################################
//
// Stub functions
// ==============
//
// ########################################################################





// Using no SSE and/or vectored memset/memcpy. These stubs prevent SSE2 code from
// linking from the LIBC.

extern "C" UINT __sse2_available = 0;

extern "C" void _cdecl _VEC_memzero () { Break (); }
extern "C" void _cdecl _VEC_memcpy () { Break (); }





// ########################################################################
//
// Functions
// =========
//
// ########################################################################





static bool Uninstall (PCTSTR const Id, bool & RebootRequired) {

  HDEVINFO DL;

  bool Success = false;

  do {

    DL = CreateDL (nullptr, DIGCF_ALLCLASSES | DIGCF_PRESENT);

    if (!DL) {

      break;

    }

    SP_DEVINFO_DATA DI;

    if (!FindDeviceById (DL, DI, Id)) {

      Error (L"Device for %s not found", Id);

      break;

    }

    wchar_t InstanceId [MAX_PATH];

    if (SetupDiGetDeviceInstanceId (DL, &DI, InstanceId, sizeof (InstanceId), nullptr)) {

      Debug (7, L"Instance ID: \"%s\"", InstanceId);

    } else {

      TranslateSetupApiError ();

      Debug (9, L"Cannot get device instance ID (%s)", LastErrorText ());

      *InstanceId = L'\0';

    }

    wchar_t Desc [MAX_PATH];

    if (
      SetupDiGetDeviceRegistryProperty (
        DL,
        &DI,
        SPDRP_DEVICEDESC,
        0,
        (BYTE *) (Desc),
        sizeof (Desc),
        0
      )
    ) {

      Debug (6, L"Desc: \"%s\"", Desc);

    } else {

      TranslateSetupApiError ();

      Debug (9, L"Cannot get device desc (%s)", LastErrorText ());

      *Desc = L'\0';

    }

    wchar_t Service [MAX_PATH];

    if (
      SetupDiGetDeviceRegistryProperty (
        DL,
        &DI,
        SPDRP_SERVICE,
        0,
        (BYTE *) (Service),
        sizeof (Service),
        0
      )
    ) {

      Debug (6, L"Service ID: \"%s\"", Service);

    } else {

      TranslateSetupApiError ();

      Debug (9, L"Cannot get device service (%s)", LastErrorText ());

      *Service = L'\0';

    }

    wchar_t Driver [256];
    wchar_t InfPath [256];

    HKEY DrvKey = SetupDiOpenDevRegKey (
      DL,
      &DI,
      DICS_FLAG_GLOBAL,
      0,
      DIREG_DRV,
      GENERIC_READ
    );

    if (DrvKey != 0) {

      if (GetRegistrySzValue (DrvKey, L"Driver", Driver, sizeof (Driver)))  {

        Debug (6, L"Driver: \"%s\"", Driver);

      } else {

        *Driver = L'\0';

      }

      if (GetRegistrySzValue (DrvKey, L"InfPath", InfPath, sizeof (InfPath)))  {

        Debug (6, L"InfPath: \"%s\"", InfPath);

      } else {

        *InfPath = L'\0';

      }

      RegCloseKey (DrvKey);
      DrvKey = 0;

    } else {

      TranslateSetupApiError ();

      Warning (L"Cannot open driver key (%s)", LastErrorText ());

      *Driver = L'\0';
      *InfPath = L'\0';

    }



    StopDeviceDriverWithRetry (DL, DI, Desc, Service);



    bool TracesLeft = false;



    if (!RemoveDevice (DL, DI, InstanceId)) {

      break;

    }

    CheckIfRebootRequired (DL, DI, RebootRequired);

    if (*InfPath) {

      Assert (wcschr (InfPath, L'\\') == nullptr);

      // SetupUninstallOEMInf removes driver package as well

#if 0

      bool RestartRecommended = false;

      if (RemoveDriverPackage (InfPath, RestartRecommended)) {

        // DriverPackageUninstall may return ERROR_FILE_NOT_FOUND but successfully
        // delete the package.

        RebootRequired |= RestartRecommended;

      }

#endif
        
      Trace (7, L"Calling SetupUninstallOEMInf (\"%s\", 0, nullptr)", InfPath);

      if (!SetupUninstallOEMInf (InfPath, 0, nullptr)) {

        TranslateSetupApiError ();

        if (GetLastError () != ERROR_FILE_NOT_FOUND && GetLastError () != ERROR_PATH_NOT_FOUND) {

          Warning (L"Cannot uninstall %s (%s)", InfPath, LastErrorText ());

          TracesLeft = true;

        }

      }

    }

    if (*Service != L'\0') {

      if (!RemoveService (Service) && GetLastError () != ERROR_SERVICE_DOES_NOT_EXIST) {

        TracesLeft = true;

      }

    }

    if (*Driver) {

      wchar_t FullPath [256];

      GetWindowsDirectory (FullPath, NumArrayElems (FullPath));

      wcscat (FullPath, L"\\system32\\drivers\\");
      wcscat (FullPath, Driver);

      if (
        !DelFile (FullPath)
        && GetLastError () != ERROR_FILE_NOT_FOUND
        && GetLastError () != ERROR_PATH_NOT_FOUND
      ) {

        TracesLeft = true;

      }

    }

    if (RebootRequired && !TracesLeft) {

      // If there are no traces left, reboot is not required

      RebootRequired = false;

    }

    Success = true;

  } OnlyOnce;

  DestroyDL (DL);

  return Success;

}





static bool Install (PCTSTR HardwareId, PCTSTR InfFile, bool &RebootRequired) {

  HDEVINFO DL = 0;
  SP_DEVINFO_DATA DI;

  wchar_t *MSZ = nullptr;

  Debug (7, L"Installing \"%s\" from \"%s\"", HardwareId, InfFile);

  bool DeviceInstanceCreated = false;
  wchar_t InstanceId [MAX_PATH];

  wcscpy (InstanceId, L"???");

  bool Success = false;

  do {

    for (UINT i = 0; i < 2; i++) {

      wchar_t Name [MAXPNAMELEN];

      if (GetPreferredDevice ((i != 0), Name)) {

        SetPreferredDevice ((i != 0), Name);

      }

    }

    GUID ClassGuid;
    wchar_t ClassName [MAX_CLASS_NAME_LEN];

    if (
      !SetupDiGetINFClass (
        InfFile,
        &ClassGuid,
        ClassName,
        NumArrayElems (ClassName),
        0
      )
    ) {

      TranslateSetupApiError ();

      Error (L"Cannot get INF class (%s)", LastErrorText ());

      break;

    }

    Debug (6, L"INF class name is \"%s\"", ClassName);

    DL = SetupDiCreateDeviceInfoList (&ClassGuid, 0);

    if (DL == INVALID_HANDLE_VALUE) {

      DL = 0;

      TranslateSetupApiError ();

      Error (L"Cannot create device info list (%s)", LastErrorText ());

      break;

    }

    wchar_t InstanceName [MAX_PATH];

    if (*InstancePath != L'\0') {

      wcscpy (InstanceName, InstancePath);

    } else {

      wcscpy (InstanceName, ClassName);

    }

    ZeroMemory (&DI, sizeof (DI));
    DI.cbSize = sizeof (DI);

    wchar_t InstanceDesc [MAX_PATH];

    wsprintf (
      InstanceDesc,
      L"\"%s\"%s",
      InstanceName, (*InstancePath != L'\0')? L"" : L" (autogenerated)"
    );

    Debug (7, L"Creating device info with instance data %s", InstanceDesc);

    if (
      !SetupDiCreateDeviceInfo (
        DL,
        InstanceName,
        &ClassGuid,
        nullptr,
        0,
        DWORD ((*InstancePath != L'\0')? 0 : DICD_GENERATE_ID),
        &DI
      )
    ) {

      TranslateSetupApiError ();

      Error (L"Cannot create device info with instance data %s (%s)", InstanceDesc, LastErrorText ());

      break;

    }

    DeviceInstanceCreated = true;

    if (!SetupDiGetDeviceInstanceId (DL, &DI, InstanceId, sizeof (InstanceId), nullptr)) {

      TranslateSetupApiError ();

      Debug (9, L"Cannot get device instance ID (%s)", LastErrorText ());

      wcscpy (InstanceId, L"?");

    }

    Debug (8, L"Created device info, node %x, instance ID \"%s\"", DI.DevInst, InstanceId);

    UINT const IdLen = static_cast <UINT> (wcslen (HardwareId));

    MSZ = new wchar_t [IdLen + 2];

    if (!MSZ) {

      Error (L"Cannot allocate %u bytes for hardware id");

      break;

    }

    memcpy (MSZ, HardwareId, StrLenToSz (IdLen));

    MSZ [IdLen] = L'\0';
    MSZ [IdLen + 1] = L'\0';

    Debug (7, L"Setting SPDRP_HARDWAREID to \"%s\"", MSZ);

    if (
      !SetupDiSetDeviceRegistryProperty (
        DL, &DI,
        SPDRP_HARDWAREID,
        PBYTE (MSZ), DWORD (StrLenToSz (IdLen + 2))
      )
    ) {

      TranslateSetupApiError ();

      Error (L"Cannot set SPDRP_HARDWAREID (%s)", LastErrorText ());

      break;

    }

    if (!SetupDiCallClassInstaller (DIF_REGISTERDEVICE, DL, &DI)) {

      TranslateSetupApiError ();

      Error (L"Cannot register device (%s)", LastErrorText ());

      break;

    }

    if (!EnableDevice (DL, DI, false)) {

      break;

    }

    Debug (
      7,
      L"Calling UpdateDriverForPlugAndPlayDevices (HwId=%s, Inf=\"%s\")",
      HardwareId, InfFile
    );

    BOOL NeedReboot;

    if (!UpdateDriverForPlugAndPlayDevices (HWND_DESKTOP, HardwareId, InfFile, 0, &NeedReboot)) {

      TranslateSetupApiError ();

      Error (L"Cannot install driver (%s)", LastErrorText ());

      break;

    }

    RebootRequired = !!NeedReboot;

    Debug (7, L"Device and driver installed, RebootRequired=%u", RebootRequired);

    if (!RebootRequired) {

      Debug (7, L"Flushing disk buffers");

      FlushVolumes ();

    }

    Debug (7, L"Enabling the device");

    if (!EnableDevice (DL, DI, true)) {

      break;

    }

    Success = true;

  } OnlyOnce;

  if (!Success && DeviceInstanceCreated) {

    Debug (7, L"Deleting device info");

    RemoveDevice (DL, DI, InstanceId);

    DeviceInstanceCreated = false;

  }

  if (MSZ != nullptr) {

    delete MSZ;
    MSZ = nullptr;

  }

  DestroyDL (DL);

  return Success;

}





//
// Entry function
// ==============
//

UINT Main (void) {

  // /opt:ref linker option removes stubs if they are not referenced in object modules.
  // __sse2_available never become nonzero so stub functions never called.

  if (__sse2_available != 0) {

    _VEC_memzero ();
    _VEC_memcpy ();

  }

#if _Debug

  GlobalDebugPrintLevel = 6;

#endif

  Debug (8, L"");
  Debug (8, ProductName L" " _US (VersionMajor) L"." _US (VersionMinor));

  UINT Res = 0;

  do {

    OSVERSIONINFO VI;

    VI.dwOSVersionInfoSize = sizeof (VI);

    Verify (GetVersionEx (&VI));

    if (VI.dwPlatformId != VER_PLATFORM_WIN32_NT || VI.dwMajorVersion < 5) {

      Error (L"OS versions before 5.0 (Windows 2000) are not supported");

      Res = 7;

      break;

    }

    PCTSTR CmdLine = GetCommandLine ();

    Debug (7, L"Command line: \"%s\"", CmdLine);  

    CmdLine = GetQuotedString (CmdLine, 0);

    if (!*CmdLine) {

      MessageBox (
        0,
        L"WDM driver manager " _US (VersionMajor) L"." _US (VersionMinor) L"\n"
        L"\n"
        L"Usage: " ProductShortName L" [<options>] <action> <device id> [<INF file>]\n"
        L"\n"
        L"Actions:\n"
        L"\n"
        L"-i - install device  and driver.\n"
        L"-r - remove device and driver.\n"
        L"\n"
        L"<device id> should be a hardware id on installation and a hardware/instance id on removal.\n"
        L"\n"
        L"<INF file> must specify device INF file path (installation only).\n"
        L"\n"
        L"Options:\n"
        L"\n"
        L"-q - no info messages.\n"
        L"-e - no error messages.\n"
        L"-h <path> - instance id path in the \"Enum\" tree, instead of auto-generated.\n"
        L"",
        L"Information",
        MB_ICONINFORMATION | MB_OK
      );

      Res = 1;

      break;

    }

    bool InstallMode = false;
    *InstancePath = L'\0';

    DWORD OptionMask = 0;

    CmdLine = SkipBlank (CmdLine);

    while (*CmdLine == L'-') {
    
      ++CmdLine;
      
      BYTE const Option = BYTE (*CmdLine++);

      if (!Option) {
      
        Error (L"Missing option letter after the dash");

        Res = 2;

        break;

      }

      if (Option < L'a' || Option > L'z') {

        Error (L"\"-%c\" must be a letter", Option);

        Res = 2;

        break;

      }

      Debug (7, L"Got option: \"%c\"", Option);

      DWORD const Mask = 1u << (Option - L'a');

      if (TestBits (OptionMask, Mask)) {

        Error (L"Option \"-%c\" specified multiple times", Option);

        Res = 2;

        break;

      }

      switch (Option) {

        case L'q': {

          NoInfoMessages = true;

          break;

        }

        case L'e': {

          NoErrorMessages = true;

          break;

        }

        case L'h': {

          CmdLine = SkipBlank (CmdLine);

          CmdLine = GetQuotedString (CmdLine, InstancePath);

          if (CmdLine != nullptr && *InstancePath != L'\0') {

            Debug (8, L"Got instance path: \"%s\"", InstancePath);

          } else {

            Error (L"No instance path");

            Res = 2;

          }

          break;

        }

        case L'i': {

          InstallMode = true;

          break;

        }

        case L'r': {

          InstallMode = false;

          break;

        }

        default: {

          Error (L"Invalid option \"-%c\"", Option);

          Res = 2;

          break;

        }

      } // switch (Option)

      if (Res != 0) {

        break;

      }

      SetBits (OptionMask, Mask);

      CmdLine = SkipBlank (CmdLine);

    } // Option loop

    if (Res != 0) {

      break;

    }

    if (
      !!TestBits (OptionMask, (1 << (L'i' - L'a')))
      == !!TestBits (OptionMask, (1 << (L'r' - L'a')))
    ) {

      Error (L"Invalid -i/-r option combination");

      Res = 2;

      break;

    }

    wchar_t HwId [256];
    wchar_t InfFile [512];

    CmdLine = GetQuotedString (CmdLine, HwId);

    if (!CmdLine) {

      Error (L"Device id not specified");

      Res = 2;

      break;

    }

    CmdLine = SkipBlank (CmdLine);

    CmdLine = GetQuotedString (CmdLine, InfFile);

    if (!CmdLine) {

      if (!InstallMode) {

        *InfFile = L'\0';

      } else {

        Error (L"INF file not specified");

        Res = 2;

        break;

      }

    }

    ProcessingStarted = true;

    wchar_t FullInfFilePath [MAX_PATH];

    if (InstallMode) {

      DWORD const Len = GetFullPathName (
        InfFile, NumArrayElems (FullInfFilePath), FullInfFilePath, 0
      );

      if (Len == 0) {

        Error (L"Cannot get full path for %s (%s)", InfFile, LastErrorText ());

        Res = 3;

        break;

      }

    }

    if (InstallMode && IsDevicePresent (HwId)) {

      Error (L"Device for %s already installed", HwId);

      Res = 6;

      break;

    }

    bool RebootRequired = false;

    bool const Success = InstallMode? (
      Install (HwId, FullInfFilePath, RebootRequired)
    ) : (
      Uninstall (HwId, RebootRequired)
    );

    if (!Success) {

      Res = 4;

      break;

    }

    wchar_t Msg [512];

    wsprintf (Msg, L"Device %sinstalled successfully", InstallMode? L"" : L"un");

    UINT MsgType;

    if (RebootRequired) {

      wcscat (Msg, L" but reboot is required");

      Res = 5;

      MsgType = MB_ICONWARNING;

    } else {

      Res = 0;

      MsgType = MB_ICONINFORMATION;

    }

    if (RebootRequired && !NoErrorMessages || !NoInfoMessages) {

      Message (MsgType | MB_OK, L"%s", Msg);

    }

  } OnlyOnce;

  Debug (8, L"Exiting (%u)", Res);

  return Res;

}





//
// Light-weight entry point if linked with /ENTRY option. Can be removed.
//

extern "C" void _cdecl WinAppEntry (void *Arg) {

  Unused (Arg);

  ExitProcess (Main ());

}





//
// Typical entry point
//

int WINAPI WinMain (
  HINSTANCE Instance,
  HINSTANCE PrevInstance,
  PSTR CmdLine,
  int CmdShow
) {

  // Make arguments L"referenced"

  Unused (Instance && PrevInstance && CmdLine && CmdShow);

  return int (Main ());

}
